#version 430 core
/*
* Copyright 2018 Bernhard Aschl and Klemens Bswirth
*/

#define MAX_LIGHTS 32
#define MAX_CENTERS 64

#define MAX_SAMPLES 160 // max samples per fragment

#define POS_INF (1.0f/0.0f)
#define NEG_INF (-1.0f/0.0f)

/*** Constants ***/
const float PI = 3.14159265358979323846264338327950288;
const float E  = 2.71828182845904523536;

/* diffuse illumination modes */
const uint ILLUMINATION_DIFFUSE_LAMBERT = 0;

/* specular illumination modes */
const uint ILLUMINATION_SPECULAR_PHONG = 0;
const uint ILLUMINATION_SPECULAR_COOK_TOR = 1;


/*** Structs ***/
struct Light 
{
    /*
    * Types:
    *   1) Spot Light -> if outerCos > -1
    *   2) Point Light -> else if direction == (0,0,0)
    *   3) Directional Light -> else
    */

    vec3 color;
	vec3 direction; // normalized
	vec3 position;
	vec3 attenuation; // (constant, linear, quadratic)
	float innerCos; // == cos(2*innerAngle)
	float outerCos; // == cos(2*outerAngle)
    float exponent; // angular attenuation exponent

    bool enabled;
};

struct Material{
	float ka;
	float kd;
	float ks;
	float alpha;
    
	float microFacetMean;
	float ior;
	bool useTex; // not supported
	bool useSpec; // not supported
};

struct Ray {
    vec3 origin;
    vec3 direction;
    vec3 inv_direction;
};

struct AABB {
    vec3 min;
    vec3 max;
};

/*** Inputs ***/

layout(pixel_center_integer) in vec4 gl_FragCoord;

/*** Uniforms ***/
layout(location=2) uniform vec3 ambientLight;
uniform vec3 diffuseColor;
uniform vec3 camera_world;
uniform Material Mat;

uniform uint lightCount;
uniform Light lights[MAX_LIGHTS];

uniform float brightness;

uniform uint diffuseIllumination;
uniform uint specularIllumination;

uniform mat4 modelViewProjMatrix;

uniform vec3 nearPlaneBL;
uniform vec3 nearPlaneBR;
uniform vec3 nearPlaneTR;
uniform float inv_scaleSqr;
uniform float isolevel;

uniform uint centersCount;
uniform vec3 centers[MAX_CENTERS];

uniform AABB aabb;
uniform vec4 viewport;

uniform float stepSize;

/*** Outputs ***/
out vec4 color;

/*** Fields ***/

vec3 position_world;
vec3 normal_world;


/*** Illumination Models ***/
vec3 illumination_diffuse(vec3 lightDirection, vec3 normal, vec3 intensity)
{
    switch(diffuseIllumination)
    {
        case ILLUMINATION_DIFFUSE_LAMBERT:
            return vec3(Mat.kd) * max(0, dot(lightDirection, normal)) * intensity;

		default:
			return vec3(0);
    }
}

vec3 illumination_specular(vec3 eyeDirection, vec3 lightDirection, vec3 intensity)
{
    const vec3 normal = normal_world;
    const vec3 reflected = -normalize(reflect(lightDirection, normal));
    const vec3 halfVec = normalize(lightDirection + eyeDirection);

    switch(specularIllumination)
    {
        case ILLUMINATION_SPECULAR_PHONG:
	        return vec3(Mat.ks) * pow(max(0, dot(reflected, eyeDirection)), Mat.alpha) * intensity;

		case ILLUMINATION_SPECULAR_COOK_TOR:
            const float mSqr = Mat.microFacetMean * Mat.microFacetMean;
            const float cosAlpha = dot(halfVec, normal);
            const float cosAlphaSqr = cosAlpha * cosAlpha;
            const float D_num = exp((cosAlphaSqr - 1)/(cosAlphaSqr * mSqr));
            const float D_denom = PI * mSqr * cosAlphaSqr * cosAlphaSqr;
            const float D = D_num / D_denom;

	        const float geom = 2 * dot(normal, halfVec) / dot(eyeDirection, halfVec);
	        const float G_mask = geom * dot(normal, eyeDirection);
	        const float G_shadow = geom * dot(normal, lightDirection);
	        const float G = min(1, min(G_mask, G_shadow));

	        const float c = max(0, dot(lightDirection, halfVec));
            const float cSqr = c * c;
	        const float g = sqrt(Mat.ior * Mat.ior + cSqr - 1);
            const float gc = g * c;

            const float gmcSqr = (g - c) * (g - c);
            const float gpcSqr = (g + c) * (g + c);
	        const float F0 = gmcSqr / 2 * gpcSqr;
            const float F1 = (gc + cSqr - 1) * (gc + cSqr - 1);
            const float F2 = (gc - cSqr + 1) * (gc - cSqr + 1);
	        const float F = F0 + F1 / F2;

            const float specDot = max(0, dot(normal, eyeDirection));
	        return vec3(Mat.ks) * D * G * F / (PI * specDot) + intensity;

		default:
			return vec3(0);
    }
}

/*** Functions ***/
void calculateLight(uint lightIndex, vec3 eyeDirection, out vec3 diffuse, out vec3 specular)
{
    vec3 lightDirection = normalize(lights[lightIndex].position - position_world);
    float attenuation = 1.0f;

    if(lights[lightIndex].outerCos > -1.0f)
    {
        // spot light
        const float dist = distance(lights[lightIndex].position, position_world);
		attenuation = 1.0f / dot(lights[lightIndex].attenuation, vec3(1, dist, dist * dist));
        
		const float spotDot = dot(-lightDirection, lights[lightIndex].direction);
        if(spotDot < lights[lightIndex].outerCos)
        {
            attenuation = 0;
        }
        else if(spotDot < lights[lightIndex].innerCos)
        {
            const float spot = smoothstep(lights[lightIndex].outerCos, lights[lightIndex].innerCos, spotDot);
            attenuation *= pow(spot, lights[lightIndex].exponent);
        }
    }
    else if(lights[lightIndex].direction == vec3(0))
    {
        // point light
        const float dist = distance(lights[lightIndex].position, position_world);
		attenuation = 1.0f / dot(lights[lightIndex].attenuation, vec3(1, dist, dist * dist));
    }
    else
    {
        // directional light
        lightDirection = -lights[lightIndex].direction;
    }

    diffuse = illumination_diffuse(lightDirection, normal_world, lights[lightIndex].color) * attenuation;
	specular = illumination_specular(eyeDirection, lightDirection, lights[lightIndex].color) * attenuation;
}

vec4 shade()
{
    const vec3 eyeDirection = normalize(camera_world - position_world);

    vec3 ambient = ambientLight * Mat.ka + brightness;
    vec3 diffuse = vec3(0);
    vec3 specular = vec3(0);

	for(int i = 0; i < lightCount; i++)
    {
        if(lights[i].enabled)
        {
            vec3 d, s;
            calculateLight(i, eyeDirection, d, s);
            diffuse += d;
            specular += s;
        }
	}

    return vec4((ambient+diffuse)*diffuseColor + specular, 1);
}

float metaball(int index, vec3 position)
{
    const vec3 radiusVec = position - centers[index];
    const float radiusSqr = dot(radiusVec, radiusVec)*inv_scaleSqr;
    return 1.f/radiusSqr;
}

float metaballs(vec3 position)
{
    float sum = 0.0f;
    for(int i = 0; i < centersCount; ++i) {
        sum += metaball(i, position);
    }
    return sum;
}

vec3 metaball_normal(int index, vec3 position)
{
    const vec3 radiusVec = position - centers[index];
    const float radiusSqr = dot(radiusVec, radiusVec)*inv_scaleSqr;
    return 2*radiusVec/(radiusSqr*radiusSqr);
}

vec3 metaballs_normal(vec3 position)
{
    vec3 sum = vec3(0);
    for(int i = 0; i < centersCount; ++i) {
        sum += metaball_normal(i, position);
    }
    return normalize(sum);
}

bool intersect_AABB(Ray ray, AABB aabb, out float t_near, out float t_far)
{
    const vec3 t1 = (aabb.min - ray.origin)*ray.inv_direction;
    const vec3 t2 = (aabb.max - ray.origin)*ray.inv_direction;

    const vec3 tmin = min(t1, t2);
    const vec3 tmax = max(t1, t2);

    t_near = max(tmin.x, max(tmin.y, tmin.z));
    t_far = min(tmax.x, min(tmax.y, tmax.z));

    return t_far >= t_near;
}

/*** Main ***/
void main() {
// raycasting partly based on http://prideout.net/blog/?p=64
	gl_FragDepth = 10e6;
    
    const vec2 normWindow = (gl_FragCoord.xy-viewport.xy) / viewport.zw;
    const vec3 nearPoint = nearPlaneBL + (nearPlaneBR - nearPlaneBL)*normWindow.x + (nearPlaneTR - nearPlaneBR)*normWindow.y;

    Ray ray;
    ray.origin = camera_world;
    ray.direction = normalize(nearPoint - camera_world);
    ray.inv_direction = 1.0f/ray.direction;

    float t_in, t_out;
    if(!intersect_AABB(ray, aabb, t_in, t_out)) {
		// don't draw outside AABB
        discard;
    }

    t_in = min(t_in, 0.0f);
    
    const vec3 p_in = ray.origin + ray.direction * t_in;
    const vec3 p_out = ray.origin + ray.direction * t_out;

    const vec3 step_world = ray.direction*stepSize;

    position_world = p_in;
    float dist = distance(p_in, p_out);
    
    for(int i = 0; i < MAX_SAMPLES && dist > 0.0f; ++i) {
        if(metaballs(position_world) > isolevel) {
            normal_world = metaballs_normal(position_world);
			const vec4 pos_proj = modelViewProjMatrix*vec4(position_world,1);
            gl_FragDepth = 0.5*(pos_proj.z/pos_proj.w + 1.0);
            color = shade();
            return;
        }

        position_world += step_world;
        dist -= stepSize;
    }

	// ray didn't hit anything
	discard;
}