#version 430 core
/*
* Copyright 2018 Bernhard Aschl and Klemens Böswirth
*/

#define MAX_LIGHTS 32

/*** Constants ***/
const float PI = 3.14159265358979323846264338327950288;
const float E  = 2.71828182845904523536;

/* diffuse illumination modes */
const uint ILLUMINATION_DIFFUSE_LAMBERT = 0;

/* specular illumination modes */
const uint ILLUMINATION_SPECULAR_PHONG = 0;
const uint ILLUMINATION_SPECULAR_COOK_TOR = 1;


/*** Structs ***/
struct Light 
{
    /*
    * Types:
    *   1) Spot Light -> if outerCos > -1
    *   2) Point Light -> else if direction == (0,0,0)
    *   3) Directional Light -> else
    */

    vec3 color;
	vec3 direction; // normalized
	vec3 position;
	vec3 attenuation; // (constant, linear, quadratic)
	float innerCos; // == cos(2*innerAngle)
	float outerCos; // == cos(2*outerAngle)
    float exponent; // angular attenuation exponent

	bool isDynamic;
    bool enabled;
};

struct Material{
	float ka;
	float kd;
	float ks;
	float alpha;
    
	float microFacetMean;
	float ior;
	bool useTex;
	bool useSpec;
	bool useNormalMap;
};

/*** Inputs ***/
in VertexData
{
    vec3 position_world;
    vec3 normal_world; // normalized
    vec2 uv_interpol;
	mat3 tangentToModelMatrix;
} vert;

/*** Uniforms ***/
layout(location=2) uniform vec3 ambientLight;
uniform vec3 diffuseColor;

uniform sampler2D colorTexture;
uniform sampler2D specTexture;
uniform sampler2D normalMap;
uniform sampler2D lightmap;

uniform vec3 camera_world;
uniform Material Mat;
uniform bool objectIsDynamic;

uniform uint lightCount;
uniform Light lights[MAX_LIGHTS];

uniform float brightness;

uniform uint diffuseIllumination;
uniform uint specularIllumination;

uniform mat4 transInvModelMatrix;

uniform bool useLightMap;

/*** Outputs ***/
out vec4 color;

/*** Illumination Models ***/
vec3 illumination_diffuse(vec3 lightDirection, vec3 normal, vec3 intensity)
{
    switch(diffuseIllumination)
    {
        case ILLUMINATION_DIFFUSE_LAMBERT:
            return vec3(Mat.kd) * max(0, dot(lightDirection, normal)) * intensity;

		default:
			return vec3(0);
    }
}

vec3 illumination_specular(vec3 ks, vec3 eyeDirection, vec3 lightDirection, vec3 normal, vec3 intensity)
{
    const vec3 reflected = -normalize(reflect(lightDirection, normal));
    const vec3 halfVec = normalize(lightDirection + eyeDirection);

    switch(specularIllumination)
    {
        case ILLUMINATION_SPECULAR_PHONG:
	        return ks * pow(max(0, dot(reflected, eyeDirection)), Mat.alpha) * intensity;

		case ILLUMINATION_SPECULAR_COOK_TOR:
            const float mSqr = Mat.microFacetMean * Mat.microFacetMean;
            const float cosAlpha = dot(halfVec, normal);
            const float cosAlphaSqr = cosAlpha * cosAlpha;
            const float D_num = exp((cosAlphaSqr - 1)/(cosAlphaSqr * mSqr));
            const float D_denom = PI * mSqr * cosAlphaSqr * cosAlphaSqr;
            const float D = D_num / D_denom;

	        const float geom = 2 * dot(normal, halfVec) / dot(eyeDirection, halfVec);
	        const float G_mask = geom * dot(normal, eyeDirection);
	        const float G_shadow = geom * dot(normal, lightDirection);
	        const float G = min(1, min(G_mask, G_shadow));

	        const float c = max(0, dot(lightDirection, halfVec));
            const float cSqr = c * c;
	        const float g = sqrt(Mat.ior * Mat.ior + cSqr - 1);
            const float gc = g * c;

            const float gmcSqr = (g - c) * (g - c);
            const float gpcSqr = (g + c) * (g + c);
	        const float F0 = gmcSqr / 2 * gpcSqr;
            const float F1 = (gc + cSqr - 1) * (gc + cSqr - 1);
            const float F2 = (gc - cSqr + 1) * (gc - cSqr + 1);
	        const float F = F0 + F1 / F2;

            const float specDot = max(0, dot(normal, eyeDirection));
	        return ks * D * G * F / (PI * specDot) + intensity;

		default:
			return vec3(0);
    }
}

/*** Functions ***/
void calculateLight(uint lightIndex, vec3 ks, vec3 eyeDirection, out vec3 diffuse, out vec3 specular, bool useLightMap)
{
    vec3 lightDirection = normalize(lights[lightIndex].position - vert.position_world);
    float attenuation = 1.0f;

    if(lights[lightIndex].outerCos > -1.0f)
    {
        // spot light
        const float dist = distance(lights[lightIndex].position, vert.position_world);
		attenuation = 1.0f / dot(lights[lightIndex].attenuation, vec3(1, dist, dist * dist));
        
		const float spotDot = dot(-lightDirection, lights[lightIndex].direction);
        if(spotDot < lights[lightIndex].outerCos)
        {
            attenuation = 0;
        }
        else if(spotDot < lights[lightIndex].innerCos)
        {
            const float spot = smoothstep(lights[lightIndex].outerCos, lights[lightIndex].innerCos, spotDot);
            attenuation *= pow(spot, lights[lightIndex].exponent);
        }
    }
    else if(lights[lightIndex].direction == vec3(0))
    {
        // point light
        const float dist = distance(lights[lightIndex].position, vert.position_world);
		attenuation = 1.0f / dot(lights[lightIndex].attenuation, vec3(1, dist, dist * dist));
    }
    else
    {
        // directional light
        lightDirection = -lights[lightIndex].direction;
    }
	vec3 normal = vert.normal_world;
	if(Mat.useNormalMap)
	{
		vec3 nmNorm = vert.tangentToModelMatrix*((2.0*texture(normalMap, vert.uv_interpol))-1.0).xyz;
		normal=normalize((transInvModelMatrix*vec4(nmNorm.xyz,1)).xyz);
	}	
	if(!useLightMap)
		diffuse = illumination_diffuse(lightDirection, normal, lights[lightIndex].color) * attenuation;
	else
		diffuse = vec3(0);
	specular = illumination_specular(ks, eyeDirection, lightDirection, normal, lights[lightIndex].color) * attenuation;
}

/*** Main ***/
void main() {
    const vec3 ks = Mat.useSpec ? vec3(Mat.ks * texture(specTexture, vert.uv_interpol).r) : vec3(Mat.ks);
    const vec3 eyeDirection = normalize(camera_world - vert.position_world);

    vec3 ambient = ambientLight * Mat.ka + brightness;
    vec3 diffuse = vec3(0);

	if(useLightMap && !objectIsDynamic)
		diffuse = texture(lightmap, vert.uv_interpol).rgb;

    vec3 specular = vec3(0);

	for(int i = 0; i < lightCount; i++)
    {
        if(lights[i].enabled)
        {
            vec3 d, s;
            calculateLight(i, ks, eyeDirection, d, s, useLightMap && !objectIsDynamic && !lights[i].isDynamic);
            diffuse += d;
            specular += s;
        }
	}
    
	if(Mat.useTex)
    {
        vec3 usedColor = texture(colorTexture, vert.uv_interpol).rgb;
        color = vec4((ambient+diffuse)*usedColor + specular, 1);
    }
    else
    {
        color = vec4((ambient+diffuse)*diffuseColor + specular, 1);
    }
}