#version 430 core
/*
* Copyright 2018 Bernhard Aschl and Klemens Bswirth
*/
#define MAX_LIGHTS 32

/*** Constants ***/
const float PI = 3.14159265358979323846264338327950288;
const float E  = 2.71828182845904523536;

/* diffuse illumination modes */
const uint ILLUMINATION_DIFFUSE_LAMBERT = 0;

/*** Structs ***/
struct Light 
{
    /*
    * Types:
    *   1) Spot Light -> if outerCos > -1
    *   2) Point Light -> else if direction == (0,0,0)
    *   3) Directional Light -> else
    */

    vec3 color;
	vec3 direction; // normalized
	vec3 position;
	vec3 attenuation; // (constant, linear, quadratic)
	float innerCos; // == cos(2*innerAngle)
	float outerCos; // == cos(2*outerAngle)
    float exponent; // angular attenuation exponent

	bool isDynamic;
    bool enabled;
};

struct Material{
	float ka;
	float kd;
	float ks;
	float alpha;
    
	float microFacetMean;
	float ior;
	bool useTex;
	bool useSpec;
	bool useNormalMap;
};

/*** Inputs ***/
in VertexData
{
    vec3 position_world;
    vec3 normal_world; // normalized
    vec2 uv_interpol;
	mat3 tangentToModelMatrix;
} vert;

/*** Uniforms ***/
uniform sampler2D normalMap;
uniform Material Mat;

uniform uint lightCount;
uniform Light lights[MAX_LIGHTS];

uniform uint diffuseIllumination;

uniform mat4 transInvModelMatrix;

/*** Outputs ***/
out vec4 color;

/*** Illumination Models ***/
vec3 illumination_diffuse(vec3 lightDirection, vec3 normal, vec3 intensity)
{
    switch(ILLUMINATION_DIFFUSE_LAMBERT)
    {
        case ILLUMINATION_DIFFUSE_LAMBERT:
            return vec3(Mat.kd) * max(0, dot(lightDirection, normal)) * intensity;

		default:
			return vec3(0);
    }
}

/*** Functions ***/
void calculateLight(uint lightIndex, out vec3 diffuse)
{
    vec3 lightDirection = normalize(lights[lightIndex].position - vert.position_world);
    float attenuation = 1.0f;

    if(lights[lightIndex].outerCos > -1.0f)
    {
        // spot light
        const float dist = distance(lights[lightIndex].position, vert.position_world);
		attenuation = 1.0f / dot(lights[lightIndex].attenuation, vec3(1, dist, dist * dist));
        
		const float spotDot = dot(-lightDirection, lights[lightIndex].direction);
        if(spotDot < lights[lightIndex].outerCos)
        {
            attenuation = 0;
        }
        else if(spotDot < lights[lightIndex].innerCos)
        {
            const float spot = smoothstep(lights[lightIndex].outerCos, lights[lightIndex].innerCos, spotDot);
            attenuation *= pow(spot, lights[lightIndex].exponent);
        }
    }
    else if(lights[lightIndex].direction == vec3(0))
    {
        // point light
        const float dist = distance(lights[lightIndex].position, vert.position_world);
		attenuation = 1.0f / dot(lights[lightIndex].attenuation, vec3(1, dist, dist * dist));
    }
    else
    {
        // directional light
        lightDirection = -lights[lightIndex].direction;
    }
	vec3 normal = vert.normal_world;
	if(Mat.useNormalMap)
	{
		vec3 nmNorm = vert.tangentToModelMatrix*((2.0*texture(normalMap, vert.uv_interpol))-1.0).xyz;
		normal=normalize((transInvModelMatrix*vec4(nmNorm.xyz,1)).xyz);
	}	

    diffuse = illumination_diffuse(lightDirection, normal, lights[lightIndex].color) * attenuation;
	
}

/*** Main ***/
void main() {
		vec3 diffuse = vec3(0);

		for(int i = 0; i < lightCount; i++)
		{
			if(lights[i].enabled && !lights[i].isDynamic)
			{
				vec3 d;
				calculateLight(i, d);
				diffuse += d;
			}
		}
    
		color = vec4(diffuse,1);
		//color = vec4(vert.uv_interpol.x, vert.uv_interpol.y, 1-vert.uv_interpol.x, 1);
		//color = vec4(1,0,0,0.5);
}